/*
  ADAU1701_DSP.cpp - Library of DSP functions for the Audiodevelopers ADAU1701 boards
  Created by Neil Davis, December, 2020.
*/

#include "Arduino.h"
#include "ADAU1701_DSP.h"
#include "Lookup_tables.h"
#include <ADAU1701_I2C.h>
#include <Bilinear.h>
#include <Shared_enums.h>


const uint8_t ADAU1701_addr = 0x34;
const byte max_TW_biquads = 5;     //maximum number of biquads used.  This constant is needed to "clear out" unused biquads
const byte max_WS_biquads = 3;
//byte BQ_coefs [5][4];   //five 28-bit values.  Will need to prepend 12 bits of zeroes to make into 40-bit data
float _coef_array[6];         //array to hold one set of biquad coefficents processed by Bilinear() and sent to coef_to_ADAU1701()


ADAU1701_I2C DSP_I2C(ADAU1701_addr);
Bilinear My_Bilinear;

ADAU1701_DSP::ADAU1701_DSP()  {}



/* Public Functions */

float ADAU1701_DSP::convert_freq_code(bool TW_true_WS_false, byte freq_code) {
  if (TW_true_WS_false == true) {
    return TW_Xover_freq_table[freq_code];    //look up the Tweeter-Woofer frequency in the table
  }
  else {
    return WS_Xover_freq_table[freq_code];    //look up the Woofer-Sub frequency in the table}
  }
}

Shared_enum::Xovers ADAU1701_DSP::convert_Xover_code(byte type_code) {
  Shared_enum::Xovers x_table[6] = {Shared_enum::none, Shared_enum::BW_1, Shared_enum::BW_3, Shared_enum::LR_2, Shared_enum::LR_4, Shared_enum::LR_8};
  x_type = x_table[type_code];     //there is a one-to-one mapping of code to enumerated type so no look-up is needed
  return x_type;
}

void ADAU1701_DSP::TW_Xover_calc(Shared_enum::BQ_spec BQ, word TW_BQ_addresses[5][4]) {
  /* This is where the crossover is calculated.  The input is a biquad_spec object, which describes the crossover filter.
      The biquad_spec has the following information:
        --filter frequency
        --the number of biquads needed to implement the crossover
        --the filter order of each biquad
        --the Q of each biquad
      The crossover_calc routine calls the biquad_load routine to calculate the biquad coefficients and puts the coefficient data in Parameter RAM
  */

  for (byte n = 0; n < BQ.biquads_needed; n++) {
    TW_biquad_load(n, BQ.Freq_center, BQ.filter_order[n], BQ.filter_q[n], TW_BQ_addresses);
  }
  fill_TW_with_none(BQ.biquads_needed, TW_BQ_addresses);                      //"zero-out" the remaining biquads
}



void ADAU1701_DSP::TW_biquad_load(byte biquad, float Freq, byte order, float q, word TW_addresses[5][4])
/*
   This routine calculates the coefficients needed to implement the desired filter by calling the coef_calc routine.
   It then converts the coefficients from floating point to 28-bit integer in the coef_to_ADAU1701 function.  It then
   loads the data into the ADAU1701 Parameter RAM using the safe load registers, so that all of the data is updated
   during the same audio sample.  The two lowpass filters are written, followed by the highpass filters
*/
{
  Shared_enum::tfilters filtertype;
  if (order == 1) filtertype = Shared_enum::One_Pole_LP; else filtertype = Shared_enum::Two_Pole_LP;

  Serial.print("Filter_enum = ");
  Serial.println(filtertype);

  My_Bilinear.coef_calc(Freq, 1.0, q, filtertype, _coef_array);
  coef_to_ADAU1701(_coef_array);
  DSP_I2C.biquad_safe_load(TW_addresses[biquad][0], ADAU1701_coefs);     //TW, Left channel, lowpass
  DSP_I2C.biquad_safe_load(TW_addresses[biquad][1], ADAU1701_coefs);     //TW, Right channel, biquad #1, lowpass


  if (order == 1) filtertype = Shared_enum::One_Pole_HP; else filtertype = Shared_enum::Two_Pole_HP;
  //Q already set


  My_Bilinear.coef_calc(Freq, 1.0, q, filtertype, _coef_array);
  coef_to_ADAU1701(_coef_array);

  DSP_I2C.biquad_safe_load(TW_addresses[biquad][2], ADAU1701_coefs);     //TW, Left channel, highpass
  DSP_I2C.biquad_safe_load(TW_addresses[biquad][3], ADAU1701_coefs);     //TW, Right channel, highpass

}

void ADAU1701_DSP::WS_Xover_calc(Shared_enum::BQ_spec BQ, word WS_BQ_addresses[3][4]) {
  /* This is where the crossover is calculated.  The input is a biquad_spec object, which describes the crossover filter.
      The biquad_spec has the following information:
        --filter frequency
        --the number of biquads needed to implement the crossover
        --the filter order of each biquad
        --the Q of each biquad
      The crossover_calc routine calls the biquad_load routine to calculate the biquad coefficients and puts the coefficient data in Parameter RAM
  */

  for (byte n = 0; n < BQ.biquads_needed; n++) {
    WS_biquad_load(n, BQ.Freq_center, BQ.filter_order[n], BQ.filter_q[n], WS_BQ_addresses);
  }
  fill_WS_with_none(BQ.biquads_needed, WS_BQ_addresses);                      //"zero-out" the remaining biquads
}

void ADAU1701_DSP::WS_biquad_load(byte biquad, float Freq, byte order, float q, word WS_addresses[5][4])
/*
   This routine calculates the coefficients needed to implement the desired filter by calling the coef_calc routine.
   It then converts the coefficients from floating point to 28-bit integer in the coef_to_ADAU1701 function.  It then
   loads the data into the ADAU1701 Parameter RAM using the safe load registers, so that all of the data is updated
   during the same audio sample.  The two lowpass filters are written, followed by the highpass filters
*/
{
  Shared_enum::tfilters filtertype;
  if (order == 1) filtertype = Shared_enum::One_Pole_LP; else filtertype = Shared_enum::Two_Pole_LP;
  My_Bilinear.coef_calc(Freq, 1.0, q, filtertype, _coef_array);
  coef_to_ADAU1701(_coef_array);

  DSP_I2C.biquad_safe_load(WS_addresses[biquad][0], ADAU1701_coefs);     //WS, Left channel, lowpass
  DSP_I2C.biquad_safe_load(WS_addresses[biquad][1], ADAU1701_coefs);     //WS, Right channel, biquad #1, lowpass

  if (order == 1) filtertype = Shared_enum::One_Pole_HP; else filtertype = Shared_enum::Two_Pole_HP;
  //Q already set
  My_Bilinear.coef_calc(Freq, 1.0, q, filtertype, _coef_array);
  coef_to_ADAU1701(_coef_array);

  DSP_I2C.biquad_safe_load(WS_addresses[biquad][2], ADAU1701_coefs);     //WS, Left channel, highpass
  DSP_I2C.biquad_safe_load(WS_addresses[biquad][3], ADAU1701_coefs);     //WS, Right channel, highpass

}

void ADAU1701_DSP::filter_calc(Shared_enum::Filter_spec Filter, word BQ_address) {

  My_Bilinear.coef_calc(Filter.Freq_center, Filter.Gain, Filter.Q, Filter.filter_type, _coef_array);
  coef_to_ADAU1701(_coef_array);
  DSP_I2C.biquad_safe_load(BQ_address, ADAU1701_coefs);
}


Shared_enum::BQ_spec ADAU1701_DSP::get_biquad_spec(float FC, Shared_enum::Xovers x_type)
/*
   This routine does a "look-up" on the enumerated crossover (X_type) to find the filter order and Q.  This "look-up" is done
   using a long case switch-case statement.  The current HCI only allows 6 different crossover types, but the switch-case statement
   actually supports 18 different crossover types, as shown below.  These can be easily added to the HCI table.  The biquad_spec
   structure is already specified to allow using 7 biquads, so any of the crossover types listed below could be implemented by just
   changing the HCI
*/
{
  struct Shared_enum::BQ_spec biquads;         //make an instance of the biquad structure called "biquads"
  biquads.Freq_center = FC;

  switch (x_type) {
    case none:
      biquads.biquads_needed = 0;
      break;

    case BW_1:
      biquads.biquads_needed = 1;
      biquads.filter_order[0] = 1;
      biquads.filter_q[0] = 1.0;
      break;

    case BW_2:
      biquads.biquads_needed = 1;
      biquads.filter_order[0] = 2;
      biquads.filter_q[0] = 0.707;
      break;

    case BW_3:
      biquads.biquads_needed = 2;
      biquads.filter_order[0] = 1;    biquads.filter_order[1] = 2;
      biquads.filter_q[0] = 0.5;      biquads.filter_q[1] = 1.0;
      break;

    case BW_4:
      biquads.biquads_needed = 2;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;
      biquads.filter_q[0] = 0.54;     biquads.filter_q[1] = 1.31;
      break;

    case BW_5:
      biquads.biquads_needed = 3;
      biquads.filter_order[0] = 1;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;
      biquads.filter_q[0] = 0.5;      biquads.filter_q[1] = 0.62;     biquads.filter_q[2] = 1.62;
      break;

    case BW_6:
      biquads.biquads_needed = 3;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;
      biquads.filter_q[0] = 0.52;     biquads.filter_q[1] = 0.71;     biquads.filter_q[2] = 1.93;
      break;

    case Bes_2:
      biquads.biquads_needed = 1;
      biquads.filter_order[0] = 2;    biquads.filter_q[0] = 0.58;
      break;

    case Bes_4:
      biquads.biquads_needed = 2;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;
      biquads.filter_q[0] = 0.52;     biquads.filter_q[1] = 0.81;
      break;

    case Bes_5:
      biquads.biquads_needed = 3;
      biquads.filter_order[0] = 1;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;
      biquads.filter_q[0] = 0.5;      biquads.filter_q[1] = 0.56;     biquads.filter_q[2] = 0.92;
      break;

    case Bes_6:
      biquads.biquads_needed = 3;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;
      biquads.filter_q[0] = 0.51;     biquads.filter_q[1] = 0.61;     biquads.filter_q[2] = 1.02;
      break;

    case LR_2:
      biquads.biquads_needed = 1;
      biquads.filter_order[0] = 2;
      biquads.filter_q[0] = 0.5;
      break;

    case LR_4:
      biquads.biquads_needed = 2;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;
      biquads.filter_q[0] = 0.707;    biquads.filter_q[1] = 0.71;
      break;

    case LR_6:
      biquads.biquads_needed = 3;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;
      biquads.filter_q[0] = 0.5;      biquads.filter_q[1] = 1.0;      biquads.filter_q[2] = 1.0;
      break;

    case LR_8:
      biquads.biquads_needed = 4;
      biquads.filter_order[0] = 2;    biquads.filter_order[1] = 2;    biquads.filter_order[2] = 2;    biquads.filter_order[3] = 2;
      biquads.filter_q[0] = 0.54;     biquads.filter_q[1] = 0.54;     biquads.filter_q[2] = 1.31;     biquads.filter_q[3] = 1.31;
      break;

    case LR_10:
      biquads.biquads_needed = 5;
      biquads.filter_order[0] = 2;  biquads.filter_order[1] = 2;  biquads.filter_order[2] = 2;
      biquads.filter_order[3] = 2;  biquads.filter_order[4] = 2;

      biquads.filter_q[0] = 0.5;    biquads.filter_q[1] = 0.62;   biquads.filter_q[2] = 0.62;
      biquads.filter_q[3] = 1.62;   biquads.filter_q[4] = 1.62;
      break;
  }
  return  biquads;
}

void ADAU1701_DSP::phase_invert(bool passthrough, word polarity_addresses[2]) {
  //set up the coefs array based on the state of the passthrough flag
  if (passthrough == true) coef_passthrough();
  if (passthrough == false) coef_minus_1();

  //now send the coefficients to the ADAU1701
  DSP_I2C.biquad_safe_load(polarity_addresses[0], ADAU1701_coefs);     //left channel tweeter, biquad #6
  DSP_I2C.biquad_safe_load(polarity_addresses[1], ADAU1701_coefs);     //right channel tweeter, biquad #6
}


/*  Private Functions  */

void ADAU1701_DSP::coef_to_ADAU1701(float coef_array[6]) {
  //convert scaled floating point filter coefficients to bytes that can be loaded into the ADAU1701 Parameter RAM
  //Notice that the ADAU1701 puts the coefficients in a different order than shown in the coef_calc routine
  const float Two_to_the_23rd = 8388608.0;
  const long bits_28 = 0x0fffffff;
  long coef_long;

  for (byte i = 0; i < 5; i++) {
    coef_array[i] = coef_array[i] * Two_to_the_23rd;
    coef_long = long(coef_array[i]);
    coef_long = coef_long & bits_28;
    ADAU1701_coefs [i][0] = lowByte((coef_long & 0x0f000000) >> 24);
    ADAU1701_coefs [i][1] = lowByte((coef_long & 0x00ff0000) >> 16);
    ADAU1701_coefs [i][2] = lowByte((coef_long & 0x0000ff00) >> 8);
    ADAU1701_coefs [i][3] = lowByte((coef_long & 0x000000ff));
  }
}


void ADAU1701_DSP::fill_TW_with_none(byte number_of_biquads, word TW_BQ_addresses[5][4])
/*      This routine simply fills in the unused biquads with "1"--that is, the filter is a pass-through.
     The coef_passthrough routine sets the coefficients to be "1", and the switch-case statement is executed for
     each biquad, until the maximum number is reached (7).
*/

{
  for (byte n = number_of_biquads; n < max_TW_biquads; n++)
  {
    Serial.print(" TW fill with none biquad_counter = ");
    Serial.println(n);
    coef_passthrough();   //Put a "1" (passthrough) in the coef array
    for (int j = 0; j < 4; j++) {
      DSP_I2C.biquad_safe_load(TW_BQ_addresses[n][j], BQ_coefs);
      Serial.print(TW_BQ_addresses[n][j]);
      Serial.print(" ");
      for (int i = 0; i < 4; i++) {
        Serial.print(BQ_coefs[j][i]);
        Serial.print(" ");
      }
      Serial.println(" ");
    }
  }
}

void ADAU1701_DSP::fill_WS_with_none(byte number_of_biquads, word WS_BQ_addresses[3][4]) {
  /*   This routine simply fills in the unused biquads with "1"--that is, the filter is a pass-through.
       The coef_passthrough routine sets the coefficients to be "1", and the switch-case statement is executed for
       each biquad, until the maximum number is reached (7).
  */

  for (byte n = number_of_biquads; n < max_WS_biquads; n++)
  {
    coef_passthrough();
    for (int j = 0; j < 4; j++) {
      DSP_I2C.biquad_safe_load(WS_BQ_addresses[n][j], BQ_coefs);
    }
  }
}


void ADAU1701_DSP::coef_passthrough() {        //Put a "1" in the biquad
  for (byte i = 0; i < max_TW_biquads; i++) {
    for (byte j = 0; j < 4; j++ ) {
      BQ_coefs [i][j] = 0x00;
    }
    BQ_coefs [0][1] = 0x80;
  }

}

void ADAU1701_DSP::coef_minus_1() {           //Put a "-1" in the biquad
  for (byte i = 0; i < max_TW_biquads; i++) {
    for (byte j = 0; j < 4; j++ ) {
      BQ_coefs [i][j] = 0x00;
    }
  }
  BQ_coefs [0][1] = 0x80;
  BQ_coefs [0][0] = 0xFF;
}
